#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>

#include "config.h"
#include "globals.h"
#include "util.h"
#include "dll.h"
#include "dllapps.h"
#include "access.h"
#include "logfun.h"
#include "readconfig.h"
#include "process.h"
#include "netio.h"

static INT   lastSignal   = -1;
static pid_t processGroup =  0;
static CHAR  *pidName     = NULL;

typedef struct cSig_s { INT   sig;
                        CHAR *nam; } cSig_t;

static cSig_t  catchedSignals[] = { { SIGTERM, "TERM" },
                                    { SIGINT,  "INT"  },
                                    { SIGCHLD, "CHLD" },
                                    { SIGHUP,  "HUP"  },
                                    { SIGUSR1, "USR1" },
                                    { SIGUSR2, "USR2" }  };

/*
	everything we need to deal with signals
*/

static VOID sigHandler (INT sigNo)
{
    signal(sigNo, (SignalHandler)sigHandler);

    lastSignal = sigNo;

    switch (sigNo) {
        case SIGUSR1:
            if (MsgLevel < 5)
                MsgLevel++;
            break;
        case SIGUSR2:
            if (MsgLevel > 0)
                MsgLevel--;
            break;
        case SIGHUP:
        case SIGINT:
        case SIGTERM:
        case SIGCHLD:
            break;
    }
}

static CHAR *findSignal (INT sigNo)
{
    INT i;

    for (i = 0; i < PARENT_SIGNALS; i++)
        if (catchedSignals[i].sig == sigNo)
            return catchedSignals[i].nam;
    return "???";
}


INT checkSignal(VOID)
{
    INT status;
    struct rusage rusage;
    pid_t child;
    dlp_t *pp;

    LOG_MSG4 (LOG_DEBUG, "Processing signal %d (%s)", 
              lastSignal, findSignal(lastSignal));

    switch (lastSignal) {
        case SIGUSR1:
        case SIGUSR2:
            LOG_MSG3 (LOG_INFO, "Message level now at %d", MsgLevel);
            break;
        case SIGHUP:
            LOG_MSG3 (LOG_INFO, "Config refresh forced");
            refreshConfig();
            break;
        case SIGINT:
            LOG_MSG3 (LOG_INFO, "Config dump to %s initiated", DumpFile);
            doDump();
            break;
        case SIGTERM:
            isAborted = 1;
            return 1;
        case SIGCHLD:
            while ((child = wait3 (&status, WNOHANG, &rusage)) > 0) {
                LOG_MSG4 (LOG_INFO, "Subprocess %d exited", child);
                if ((pp = findPid (procListH, child)) == NULL) {
                    LOG_MSG1 (LOG_ERR, "Subprocess %d not found", child);
                }
                else {
                    if (pp->ppid > 0 && pp->signal > 0) {
                        if (pp->ppid == getpid()) {
                            LOG_MSG1 (LOG_ERR, "Child died - suicide %d (%s)",
                                      pp->signal, findSignal(pp->signal));
                            raise (pp->signal); 
                        }
                        else if (kill (pp->ppid, pp->signal)) {
                            LOG_MSG1 (LOG_ERR, "Cannot send signal %d (%s) to "
                                      "parent process %d: %m", 
                                      pp->signal, findSignal(pp->signal), 
                                      pp->ppid);
                        }
                    }
                    removeElement ((dll_t *)pp, DLL_FREE);
                }
                dumpProcList();
            }
            break;
    }
    lastSignal = -1;
    return 0;
}


INT signalSetup(INT what)
{
    INT i;

    for (i = 0; i < what; i++) {
        if(signal (catchedSignals[i].sig, 
                   (SignalHandler)sigHandler) == SIG_ERR) {
            LOG_MSG1 (LOG_ERR, "Cannot set up handler for signal %s: %m",
                      catchedSignals[i].nam);
            return 1;
        }
    }
    LOG_MSG4 (LOG_DEBUG, "%d signals catched", what);
    return 0;
}
       
INT signalReset(INT what)
{
    INT i;

    for (i = 0; i < what; i++) {
        if(signal (catchedSignals[i].sig, SIG_DFL) == SIG_ERR) {
            LOG_MSG1 (LOG_ERR, "Cannot reset handler for signal %s: %m",
                      catchedSignals[i].nam);
            return 1;
        }
    }
    LOG_MSG4 (LOG_DEBUG, "%d signals released", what);
    return 0;
}

/*
	clean up at exit of master process
*/

VOID processRundown(VOID)
{
    INT status;
    struct rusage *rusage;
    
    if (pidName != NULL) {
        if (remove (pidName))
            LOG_MSG1 (LOG_ERR, "Cannot remove PID file '%s': %m", pidName);
        free (pidName);
    }
    if (isAborted) {
        LOG_MSG2 (LOG_WARNING, "terminated due to TERM signal");
        killpg (processGroup, SIGTERM);
    }
    while ((INT)wait3(&status, 0, rusage) > 0);
}


/*
	put a stamp into a pid file
*/

static INT doPidFile(VOID)
{
    INT  fd, rtc;
    CHAR pidNum[32];

    sprintf(pidNum, "%d", getpid());

    if ((pidName = (CHAR *)malloc(strlen(PIDFILE)+strlen(pidNum)+1)) == NULL) {
        errNoDynmem("PID file");
        return 1;
    }

    sprintf(pidName, PIDFILE, pidNum);

    if ((fd = open (pidName, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0) {
        strcat(pidNum, "\n");
        rtc = write(fd, pidNum, strlen(pidNum));
        close (fd);
        if (rtc == strlen(pidNum))
            return 0;
    }
      
    remove (pidName);
    free (pidName);
    LOG_MSG1 (LOG_ERR, "Error creating PID file: %m");
    return 1;
}

/*
	take all actions necessary to becom a daemon
*/

INT becomeDaemon ()
{
    INT tty, cfgFd;

    if (isDaemon)
        return 0;

    if (doFork(0))
        exit (0);

    if ((INT)(processGroup = setpgrp()) == -1) {
        LOG_MSG1 (LOG_ERR, "Cannot change process group: %m");
        return 1;
    }
    processGroup = getpgrp();

    cfgFd = getConfigFd();

    for (tty = 0; tty < 3; tty++)
        if ((tty != cfgFd) && !isaSocketFd(tty))
            close (tty);

#ifdef USE_TIOCNOTTY
    if ((tty = open("dev/tty", O_RDWR)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot access controlling terminal: %m");
        return 1;
    }   
#ifdef USE_SETOWN
    if (fcntl (tty, F_SETOWN, getpid())) {
        LOG_MSG1 (LOG_ERR, "Cannot set tty owner: %m");
        close (tty);
        return 1;
    }
#endif /* USE_SETOWN */   
    if (ioctl (tty, TIOCNOTTY, NULL)) {
        LOG_MSG1 (LOG_ERR, "Cannot detach from terminal: %m");
        close (tty);
        return 1;
    }
    close (tty);
#endif /* USE_TIOCNOTTY */

    isDaemon = 1;

    if (doPidFile() || signalSetup(1)) {
        return 1;
    }

    LOG_MSG4 (LOG_DEBUG, "Detached PID %d, PID-file '%s'",
              (INT)processGroup, pidName);

    return 0;
}


/*
	create an new child process
*/

pid_t doFork(INT dolog)
{
    pid_t pid;

    if ((pid = fork()) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot fork: %m - exiting");
        exit (1);
    }
    if (pid && dolog)
        LOG_MSG4 (LOG_DEBUG, "Forked child PID %d", (INT)pid);

    return pid;
}

/*
	switch to runtime UID/GID and set new 'root' if desired
*/

VOID changeIdent(VOID)
{
    if (!spMode && changeRoot && (RootDirectory != NULL)) {
        if (chroot(RootDirectory)) {
            LOG_MSG1 (LOG_ERR, "Cannot change to root '%s' : %m - exiting",
                      RootDirectory);
            exit (1);
        }
    }

    setegid(childGid);
    seteuid(childUid);

    if ((geteuid() != childUid) || (getegid() != childGid)) {
        LOG_MSG1 (LOG_ERR, "Cannot change to UID/GID %d/%d : %m - exiting",
                  (INT)childUid, (INT)childGid);
        exit (1);
    } 
    else {
        LOG_MSG4 (LOG_DEBUG, "Changed to UID/GID %d/%d",
                  (INT)childUid, (INT)childGid);
    } 
}

/*
	reset from runtime to real UID/GID
*/

VOID resetIdent(VOID)
{
    seteuid(getuid());
    setegid(getgid());

    if ((geteuid() != getuid()) || (getegid() != getgid())) {
        LOG_MSG1 (LOG_ERR, "Cannot reset UID/GID to %d/%d : %m - exiting",
                  (INT)getuid(), (INT)getgid());
        exit (1);
    }
}


/*
	start a filter as a slave process
	and insert it into the i/o channel chain
*/

INT forkFilter (CHAR *filter, INT mode, INT *ds)
{
#define CMAXARGS 64

    INT dspair[2], slave_in, slave_out, sts, n, nclose;
    INT cargc, rtc;
    CHAR *cargv[CMAXARGS], *ctx, *path, *ptr;

    if (pipe(dspair)) {
        LOG_MSG1 (LOG_ERR, "Cannot create pipe to filter: %m");
        return -1;
    }

/*
	ds points to the file descriptor currently connected
	to the file on disk. May be reading (RT_RRQ) or writing (RT_WRQ)
*/
    switch (mode) {
        case RT_RRQ:
            slave_in  = *ds;		/* filter takes input from file */
            slave_out = dspair[1];	/* and writes to pipe           */
            *ds       = dspair[0];	/* fileio() now reads from pipe */
            nclose    = dspair[1];	/* to be closed by parent	*/
            break;
        case RT_WRQ:
            slave_in  = dspair[0];	/* filter takes input from pipe */
            slave_out = *ds;		/* and writes to disk           */
            *ds       = dspair[1];	/* fileio() now writes to pipe  */
            nclose    = dspair[0];	/* to be closed by parent	*/
            break;
        default:
            LOG_MSG1 (LOG_ERR, 
                      "Illegal argument '%d' in call to forkFilter", mode);
            return -1;
            break;
    }

    if (!(sts = fork())) {

/**/    dup2 (slave_in,  0);	/* map stdin/stdout to filter i/o */
/**/    dup2 (slave_out, 1);
/**/
/**/    (VOID)cutConfigLine (filter, &cargc, cargv, (CMAXARGS-1), &ctx);
/**/
/**/    path = cargv[0];		/* expect full path in 1st arg. */
/**/    cargv[0] = baseName(path);	/* replace by short name	*/
/**/    cargv[cargc] = NULL;		/* and append a NULL argument	*/
/**/
/**/    for (n=2; n<32; n++)	/* close all remaining channels   */
/**/        close (n);
/**/
/**/    execv (path, cargv);		/* finally, start the filter	*/
/**/
/**/	/* we should never get here */
/**/
/**/    fprintf (stderr, "Cannot execv \"%s\": %m\n", path);
/**/    exit (1);
    }

/*
	parent process continues here
*/

    close (nclose);	/* close pipe side given to child */

    if (sts < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot fork filter: %m");
        close (*ds);
        return -1;
    }
    ptr = baseName(filter);

    if ((rtc = addPid(procListH, 0, sts, getpid(), SIGTERM, 0, 0, NULL)))
        LOG_MSG1 (LOG_ERR, "Cannot add filter '%s' to process list", ptr); 
    dumpProcList();

    LOG_MSG4 (LOG_DEBUG, "Forked filter '%s' PID %d", ptr, (INT)sts); 

    if (ptr != NULL)
        free (ptr);

    return rtc;
}
